#include "getCoverage.h"

void setPath(char * result, 
             const char * output_prefix,
             const char * prefix,
             const char * contigid)
{
  if (output_prefix)
  {
    if (output_prefix[strlen(output_prefix)-1] == '/')
    {
      //split mode
      strcpy(result, output_prefix);
      strcat(result, contigid);
    }
    else
    {
      strcpy(result, output_prefix);
    }
  }
  else
  {
    strcpy(result, prefix);
  }
}

// put the coordinates into normalized (start,end) form
void standardizeCoordinates(int * pos, int * end, 
                            int radius, int contig_len, int contig_gaps,
                            int * ps_start, int * ps_end)
{
  int s_start = 0;
  int s_end = contig_len;

  // Put coordinates into standardized form (s_start,s_end)
  if (end)
  {
    if (*end >= 0)
    {
      // *end is positive, compute from pos or 0
      s_start = pos ? *pos : 0;
      s_end   = *end;
    }
    else
    {
      // *end is negative, compute from pos or end 
      s_end = pos ? *pos : contig_len - contig_gaps;
      s_start = s_end + *end + 1; 
    }
  }
  else if (pos)
  {
    if (*pos < 0)
    {
      // Negative x provided, go from -*pos to end
      s_start = - *pos;
      s_end = contig_len;
    }
    else if (radius >= 0)
    {
      // radius coordinates
      s_start = *pos - radius;
      s_end   = *pos + radius;
    }
    else
    {
      s_start = 1;
      s_end = *pos;
    }
  }

  *ps_start = s_start;
  *ps_end   = s_end;
}



// Converts command line parameters into gapped start,end coordinates
// needed to loop through contig
void convertCoordinates(struct CONTIG_SEQUENCE *contig,
                        int * ungapped_pos, int * gapped_pos, int radius,
                        int * ungapped_end, int * gapped_end,
                        const char * seqname_range,
                        Coordinates * coords)
{
  int start = 0;
  int end = 0;
  int ugInd = 0;

  char buffer[4][32];
  if (ungapped_pos) { snprintf(buffer[0], 32, "%d", *ungapped_pos); }
  if (gapped_pos)   { snprintf(buffer[1], 32, "%d", *gapped_pos);   }
  if (ungapped_end) { snprintf(buffer[2], 32, "%d", *ungapped_end); }
  if (gapped_end)   { snprintf(buffer[3], 32, "%d", *gapped_end);   }

  logMsg(DEBUG_HIGH, "convertCoordinates",
         "ug_pos=%s, g_pos=%s, radius=%d, ug_end=%s, g_end=%s\n",
         ungapped_pos ? buffer[0] : "NULL", 
         gapped_pos   ? buffer[1] : "NULL", 
         radius, 
         ungapped_end ? buffer[2] : "NULL", 
         gapped_end   ? buffer[3] : "NULL");

  if (seqname_range)
  {
    logMsg(DEBUG_HIGH, "convertCoordinates",
           "seqname_range specified: setting start=0, ugInd=0, end=contig->len");

    coords->start = 0;
    coords->ugInd = 0;
    coords->end = contig->len;

    return;
  }

  // The main loop is based on 0-based gapped coordinates and a 0-based
  // ungapped offset. To complicate things, coordinates can be specified 
  // in (x,y), (-x), (x,-y), or (x,z) forms as both 1-based gapped
  // and ungapped numbers on the command line.

  if (gapped_end || gapped_pos)
  {
    int i;

    standardizeCoordinates(gapped_pos, gapped_end, 
                           radius, contig->len, 0,
                           &start, &end);
    logMsg(DEBUG_MEDIUM, "convertCoordinates",
           "start=%d, end=%d\n", start, end);

    // Shift to 0-based start position and validate
    start--; 
    start = (start >= 0 && start <= contig->len) ? start : 0;

    // Find Ungapped Index
    ugInd = 0;
    for (i = 0; i < start; i++) if (contig->seq[i] != '-') ugInd++;
    
    // Validate end coordinate 
    end = (end <= contig->len) ? end : contig->len;
  }
  else if (ungapped_end || ungapped_pos)
  {
    int ug_start;
    int ug_end;

    standardizeCoordinates(ungapped_pos, ungapped_end, 
                           radius, contig->len, contig->numOfGaps,
                           &ug_start, &ug_end);
    
    logMsg(DEBUG_MEDIUM, "convertCoordinates",
           "start=%d, end=%d\n", ug_start, ug_end);

    // Validate starting position
    ug_start = (ug_start > 0 && ug_start <= contig->len) ? ug_start : 1;

    // End position is ungapped distance from start
    ug_end = (ug_end <= contig->len - contig->numOfGaps) ? ug_end :
                                            contig->len - contig->numOfGaps;
    ug_end = (ug_start < ug_end) ? ug_end - ug_start : 0;

    // Ungapped offset is easy since we start with ungapped numbers
    ugInd = ug_start-1;

    // Find gapped coordinates for ug_start
    for (start = 0; ug_start && start < contig->len; start++)
    {
      if (contig->seq[start] != '-') ug_start--;
    }

    // Resume counting to fing gapped end position
    ug_end++; //include all gaps at end
    for (end = start; ug_end && end < contig->len; end++)
    {
      if (contig->seq[end] != '-') ug_end--;
    }

    // Step back to proper position if not at end of assembly 
    if (!ug_end) end--;
    start--;

    // Validate coordinates 
    start = (start >= 0 && start <= contig->len) ? start : 0;
    end = (end <= contig->len) ? end : contig->len;
  }
  else
  {
    // Invalid coordinates specified, run through the entire contig
    
    start = 0;
    end   = contig->len;

    ugInd = start;
  }

  logMsg(DEBUG_MEDIUM, "convertCoordinates (final)",
         "start = %d, end = %d, ugInd=%d\n", start, end, ugInd);

  if (start > end)
  {
    printError("convertCoordinate", DEBUG_ERROR,
               "Invalid Range Specified [%d, %d]", start, end);
    exit(1); //invalid range
  }

  coords->start = start;
  coords->ugInd = ugInd;
  coords->end = end;
}


/*
  printTotalCov()
  
  print total coverage and seq_id index table
  
*/
void printTotalCov(struct CONTIG_SEQUENCE *contig,
                   struct CONTIG_SEQUENCE **bcseq,
                   tilingData *td, 
                   int allowGaps,
                   Coordinates * coords,
                   const char * output_prefix,
                   const char * prefix,
                   int print_asmbl_id,
                   int print_offset,
                   int showReadId,
                   int doGapped)
{
  static int openForWriting = 1;
  static int readIdOffset = 0;  // Assign each read a unique id when processing multiple contigs
  const char * openmode = print_asmbl_id && !openForWriting ? "a" : "w";

  char tcovFile   [FILENAME_MAX];
  char idTblFile  [FILENAME_MAX];
  char offsetFile [FILENAME_MAX];

  FILE *f, *f_offset = NULL;

  int tilingEnd = -1;
  int tilingStart = -1;
  int tilingpos;

  libSlice_Slice s;

  int i;

  // If the output is being split, reset the counter
  if (!print_asmbl_id) { readIdOffset = 0; }

  if (output_prefix && !strcmp(output_prefix, "-"))
  {
    f = stdout;

    if (print_offset)
    {
      f_offset = stdout;
    }
  }
  else
  {
    setPath(tcovFile, output_prefix, prefix, contig->id);

    // Set the tcov_offset filename
    strcpy(offsetFile,tcovFile);
    strcat(offsetFile,".tcov_offset");
    
    // Set the idTbl filename
    strcpy(idTblFile, tcovFile);
    strcat(idTblFile, ".idTbl");

    // Set the suffix
    strcat(tcovFile, ".tcov");

    f = fOpen(tcovFile, openmode);

    printInfo("Writing coverage data for %s to %s\n", contig->id, tcovFile);

    if (print_offset)
    {
      f_offset = fOpen(offsetFile, openmode);

      printInfo("Writing offset table for %s to %s\n", contig->id, offsetFile);
    }
  }
  
  for (i = coords->start; i < coords->end; i++)
  {
    char c = contig->seq[i];

    if (i > tilingEnd)
    {
      tilingStart = i;
      tilingEnd = loadTilingChunk(i, td, bcseq, contig, 1);

      s.qv = pool_malloc(td->maxDepth);
      s.rc = pool_malloc(td->maxDepth);
      s.bc = pool_malloc(td->maxDepth+1);
    }

    tilingpos = i - tilingStart;
    
    if (allowGaps || c != '-')
    {
      int j;
      libSlice_Consensus consensusResults;

      int coordinate = coords->ugInd + coords->coordinateOffset;
      if (c == '-' && coordinate > 0) coordinate--;

      // Fill out the slice
      s.c    = c;
      s.dcov = td->dcov[tilingpos];
      s.bc[s.dcov]='\0';

      for(j = 0; j < s.dcov; j++)
      {
        unsigned int readid = td->idcov[tilingpos][j];
        int offset = bcseq[readid]->offset;
        s.rc[j] = bcseq[readid]->rc;

        s.bc[j] = bcseq[readid]->seq[(i-offset)%contig->len];
        s.qv[j] = bcseq[readid]->qvseq[(i-offset)%contig->len];
      }
          
      libSlice_getConsensus(&s, &consensusResults, NULL, 0);
      
      if (print_offset)
      {
        fprintf(f_offset, "%lld\n", ftello(f));
      }

      if (print_asmbl_id) fprintf(f,"%s ", contig->id);

      if (doGapped)
      {
        fprintf(f, "%d ", i+coords->coordinateOffset);
      }


      fprintf(f, "%d %c %d %s ",
              coordinate, c, consensusResults.qvConsensus, s.bc);

      if (s.dcov)
      {
        for(j = 0; j < s.dcov; j++) 
        {
          fprintf(f,"%d%c", s.qv[j], (j < s.dcov-1) ? ':' : ' ');
        }

        for(j = 0; j < s.dcov; j++) 
        {
          fprintf(f, "%d%c",td->idcov[tilingpos][j] + readIdOffset,
                 (j < s.dcov-1)? ':' : '\n');
        }
      }
      else
      {
        fprintf(f,"\n");
      }

      if (c != '-') coords->ugInd++;
    }
  }

  if (f_offset && f_offset != stdout) { fclose(f_offset); }

  // Open the idTbl
  if (f != stdout)
  {
    fclose(f);
    f = fOpen(idTblFile, openmode);

    printInfo("Writing id table for %s to %s\n", contig->id, idTblFile);
  }

  openForWriting = 0;

  // Print the id table
  for (i = 0; i < td->numOfSeqs; i++)
  {
    if (showReadId) { fprintf(f, "%d ", i+readIdOffset); }
    fprintf(f, "%s %d\n", bcseq[i]->id, bcseq[i]->rc);
  }

  if (f != stdout) { fclose(f); }
  readIdOffset += td->numOfSeqs;
}


/*
  printStats()
  
  print basic statistics for each slice of the coverage:
  ungapped and gapped position of the slice
  consensus base
  depth of coverage
  consensus quality value (computed using Churchill-Waterman algorithm)
  consensus quality class 
  [base coverage]
  [calculated consensus base from libSlice]
  
*/
void printStats(struct CONTIG_SEQUENCE *contig,
                struct CONTIG_SEQUENCE **bcseq,
                tilingData *td, 
                int allowGaps,
                Coordinates * coords,
                const char * output_prefix,
                const char * prefix,
                int print_asmbl_id,
                int doBasesCoverage,
                int doCalcConsensus)
{
  static int openForWriting = 1;
  const char * openmode = print_asmbl_id && !openForWriting ? "a" : "w";

  char statsFile[FILENAME_MAX];

  int i;
  FILE *f;

  int tilingEnd = -1;
  int tilingStart = -1;
  int tilingpos;

  libSlice_Slice s;

  if (output_prefix && !strcmp(output_prefix, "-"))
  {
    f = stdout;
  }
  else
  {
    setPath(statsFile, output_prefix, prefix, contig->id);

    // Set the suffix
    strcat(statsFile, ".tcovStats");

    f = fOpen(statsFile, openmode);
    openForWriting = 0;

    printInfo("Writing stats for %s to %s\n", contig->id, statsFile);
  }
  
  for (i = coords->start; i < coords->end; i++)
  {
    char c = contig->seq[i];

    if (i > tilingEnd)
    {
      tilingStart = i;
      tilingEnd = loadTilingChunk(i, td, bcseq, contig, 1);

      s.qv = pool_malloc(td->maxDepth);
      s.rc = pool_malloc(td->maxDepth);
      s.bc = pool_malloc(td->maxDepth+1);
    }

    tilingpos = i - tilingStart;

    if (allowGaps || c != '-')
    {
      int j;
      int basesCoverage = 0;
      char qc;
      libSlice_Consensus consensusResults;

      int coordinate = coords->ugInd + coords->coordinateOffset;
      if (c == '-' && coordinate > 0) coordinate--;

      // Fill out the slice
      s.dcov = td->dcov[tilingpos];
      s.c    = c;

      s.bc[s.dcov]='\0';
      for(j = 0; j < s.dcov; j++)
      {
        unsigned int readid = td->idcov[tilingpos][j];
        int offset = bcseq[readid]->offset;

        s.rc[j] = bcseq[readid]->rc;
        s.qv[j] = bcseq[readid]->qvseq[(i-offset)%contig->len];
        s.bc[j] = bcseq[readid]->seq[(i-offset)%contig->len];

        if (s.bc[j] != '-') basesCoverage++;
      }
          
      libSlice_getConsensusParam(&s, &consensusResults, NULL, 0, 
                                 (doCalcConsensus > 1) ? 1 : 0);

      if (doCalcConsensus == 3)
      {
        libSlice_updateAmbiguity(&s, &consensusResults, 0);
        consensusResults.consensus = 
              libSlice_convertAmbiguityFlags(consensusResults.ambiguityFlags); 
      }
      else if (doCalcConsensus == 4)
      {
        libSlice_updateAmbiguityConic(&s, &consensusResults, 0);
        consensusResults.consensus = 
              libSlice_convertAmbiguityFlags(consensusResults.ambiguityFlags); 
      }

      if (doCalcConsensus && s.dcov)
      {
        // Use the new consensus for qc calculation
        s.c = consensusResults.consensus;
      }
      else
      {
        consensusResults.consensus = c;
      }

      qc = libSlice_getConsQC(&s, 0);

      // Print the stats
      if (print_asmbl_id) fprintf(f, "%s ", contig->id);
      fprintf(f, "%d %d %c %d %u %d",
              coordinate, 
              i+coords->coordinateOffset,
              c, 
              s.dcov,
              consensusResults.qvConsensus,
              qc);

      if (doBasesCoverage) fprintf(f, " %d", basesCoverage);
      if (doCalcConsensus) fprintf(f, " %c", consensusResults.consensus);
      fprintf(f, "\n");

      if (c != '-') coords->ugInd++;
    }
  }
  
  if (f != stdout) fclose(f);
}



/*
  printCov()
  
  print the coverage incorporating seqID info and strendedness into the output
*/
void printCov(struct CONTIG_SEQUENCE *contig,
              struct CONTIG_SEQUENCE **bcseq,
              tilingData *td, 
              int allowGaps,
              Coordinates * coords,
              const char * output_prefix)
{
  int i,j;
  int numOfSeqs = td->numOfSeqs;

  int tilingEnd = -1;
  int tilingStart = -1;
  int tilingpos;

  int maxval = numOfSeqs;
  int numdigits = 1;
  int divide = 1;

  // Because the bcseq has only been loaded with sequences that tiling the range of interest,
  // and those sequences have been sorted by their offset, it is safe to use their readid
  // as their position in ch and qv

  char * qv = (char *) newmem2(numOfSeqs,   sizeof(char), __FILE__,__LINE__);
  char * ch = (char *) newmem2(numOfSeqs+1, sizeof(char),
                               __FILE__,__LINE__); // Null Terminated

  // Assume stdout unless there is a prefix
  FILE * f = stdout;
  if (output_prefix && output_prefix != "-")
  {
    char buffer[FILENAME_MAX];
    strcpy(buffer, output_prefix);

    if (output_prefix[strlen(output_prefix)-1] == '/')
    {
      strcat(buffer, contig->id);
    }

    strcat(buffer, ".cov");
    f = fOpen(buffer, "w");
    
    printInfo("Writing coverage data for %s to %s\n", contig->id, buffer);
  }

  // Initize null-terminator
  ch[numOfSeqs] = '\0';

  fprintf(f, "asmbl_id: %s\n", contig->id);

  while (maxval >= 10)
  {
    divide *= 10;
    maxval /= 10;
    numdigits++;
  }

  for (; numdigits > 0; numdigits--, divide /= 10)
  {
    if (numdigits == 1) { fprintf(f, "Ungapped  Consensus  "); }
    else                { fprintf(f, "                     "); }

    // Number the nucs with read numbers
    for (j=0; j < numOfSeqs; j++)
    {
      int curval = (j/divide) % 10;
      fprintf(f, "%d", curval);
    }

    fprintf(f, " ");

    // Number the quals with read numbers
    for (j=0; j < numOfSeqs; j++)
    {
      int curval = (j/divide) % 10;
      fprintf(f, "  %d", curval);
    }

    fprintf(f, "\n");
  }
      
  for (i = coords->start; i < coords->end; i++)
  {
    char c = contig->seq[i];

    if (i > tilingEnd)
    {
      tilingStart = i;
      tilingEnd = loadTilingChunk(i, td, bcseq, contig, 1);
    }

    tilingpos = i - tilingStart;

    if (c != '-' || allowGaps)
    {
      int j;
      int coordinate;

      // filling the columns with '*' characters
      for (j = 0; j < numOfSeqs; j++)
      {
        ch[j] = '.';
        qv[j] = -1;
      }
      
      // Set data for columns that have reads
      for (j = 0; j < td->dcov[tilingpos]; j++)
      {
        unsigned int readid = td->idcov[tilingpos][j];
        int offset = bcseq[readid]->offset;

        ch[readid] = bcseq[readid]->seq[(i-offset)%contig->len];
        qv[readid] = bcseq[readid]->qvseq[(i-offset)%contig->len];
      }

      coordinate = coords->ugInd + coords->coordinateOffset;
      if (c == '-' && coordinate > 0) coordinate--;

      // Print coverage data
      fprintf(f, "%8d  %9c  %s (", coordinate, c, ch);

      for (j = 0; j < numOfSeqs; j++)
      {
        if (j) fprintf(f, " ");

        if (qv[j] > -1) fprintf(f, "%2d",qv[j]);
        else            fprintf(f, " .");
      }

      fprintf(f, ")\n");

      if (c != '-') coords->ugInd++;
    }
  }

  fprintf(f, "\nSequence IDs  RC  Range\n");
  for(j = 0; j < numOfSeqs; j++) 
  {
    fprintf(f, "%d: %s %d <%d %d>\n",
            j,
            bcseq[j]->id,
            bcseq[j]->rc,
            bcseq[j]->asm_l,
            bcseq[j]->asm_r
            );
  }
  fprintf(f, "\n");

  free(ch);
  free(qv);

  if (f != stdout) fclose(f);
}



/*
  printXML()
  
  print the Slice XML of a range of slices
*/
void printXML(struct CONTIG_SEQUENCE *contig,
              struct CONTIG_SEQUENCE **bcseq,
              tilingData * td,
              int allowGaps,
              Coordinates * coords,
              const char * output_prefix)
{
  int gInd;  // gapped index
  int i;

  FILE * f = stdout;
  int tilingEnd = -1;
  int tilingStart = -1;
  int tilingpos;

  // Assume stdout unless there is a prefix and its not "-"
  if (output_prefix && strcmp(output_prefix, "-"))
  {
    char buffer[FILENAME_MAX];
    strcpy(buffer, output_prefix);

    if (output_prefix[strlen(output_prefix)-1] == '/')
    {
      strcat(buffer, contig->id);
    }

    strcat(buffer, ".slice");
    f = fOpen(buffer, "w");
    
    printInfo("Writing Slice XML for %s to %s\n", contig->id, buffer);
  }

  // Write XML Header

  fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  fprintf(f, "<!DOCTYPE Request SYSTEM \"" SLICE_DTD "\">\n");
  fprintf(f, "<Request Type=\"Consensus\" Host=\"getCoverage\" User=\"getCoverage\" Option=\"All\">\n");
  fprintf(f, "  <ReadCollection>\n");

  for (i = 0; i < td->numOfSeqs; i++)
  {
    fprintf(f, "    <Read Id=\"%d\" Seqname=\"%s\" Dir=\"%d\" Chem=\"\">\n"\
               "      <Clr Left=\"%d\" Right=\"%d\" />\n    </Read>\n",
            i, 
            bcseq[i]->id,
            bcseq[i]->rc, 
            bcseq[i]->seq_l,
            bcseq[i]->seq_r);
  }

  fprintf(f, "  </ReadCollection>\n  <SliceRange Index=\"%d\" Gindex=\"%d\">\n", 
          coords->ugInd+coords->coordinateOffset, 
          coords->start+coords->coordinateOffset);

  coords->ugInd = 0;
  gInd = 0;

  // Loop through each slice in range
  for (i = coords->start; i < coords->end; i++)
  {
    char c = contig->seq[i];

    if (i > tilingEnd)
    {
      tilingStart = i;
      tilingEnd = loadTilingChunk(i, td, bcseq, contig, 1);
    }

    tilingpos = i - tilingStart;

    if (c != '-' || allowGaps)
    {
      int j;
      int depth = td->dcov[tilingpos];
      int coordinate = coords->ugInd;
      unsigned int readid;
      int offset;

      if (c == '-' && coordinate > 0) coordinate--;

      // Write XML for Slice
      fprintf(f, "    <Slice Index=\"%d\" Gindex=\"%d\" Existing=\"%c\">\n"\
                 "      <Nuc>",
             coordinate+coords->coordinateOffset, 
             gInd+coords->coordinateOffset, 
             c);

      for (j = 0; j < depth; j++) 
      {
        readid = td->idcov[tilingpos][j];
        offset = bcseq[readid]->offset;
        fprintf(f, "%c", bcseq[readid]->seq[(i-offset)%contig->len]);
      }
      fprintf(f, "</Nuc>\n      <Qualval>");
      for (j = 0; j < depth; j++)
      {
        readid = td->idcov[tilingpos][j];
        offset = bcseq[readid]->offset;

        if (j) fprintf(f, " ");

        fprintf(f, "%d", bcseq[readid]->qvseq[(i-offset)%contig->len]);
      }
      fprintf(f, "</Qualval>\n      <ReadID>");
      for (j = 0; j < depth; j++)
      {
        if (j) fprintf(f, " ");
        fprintf(f, "%d", td->idcov[tilingpos][j]);
      }
      fprintf(f, "</ReadID>\n    </Slice>\n");

      if (c != '-') coords->ugInd++;
    }
    
    gInd++;
  }

  fprintf(f, "  </SliceRange>\n</Request>\n");

  fflush(f);
  if (f != stdout) fclose(f);
}


/*
  printVertical
  
  print vertical output suitable for Realigner
  
*/
void printVertical(struct CONTIG_SEQUENCE *contig,
                   struct CONTIG_SEQUENCE **bcseq,
                   tilingData *td, 
                   Coordinates * coords,
                   const char * output_prefix,
                   const char * prefix)
{
  char tcovFile   [FILENAME_MAX];

  int tilingEnd = -1;
  int tilingStart = -1;
  int tilingpos;

  FILE *f = NULL;

  int MAXDEPTH = 200; // Fit within Realigner's constraints

  int i, j, k;
  char columnStatus [MAXDEPTH];
  char columnDisplay [MAXDEPTH];
  int  columnID [MAXDEPTH];

  if (output_prefix && !strcmp(output_prefix, "-"))
  {
    f = stdout;
  }
  else
  {
    setPath(tcovFile, output_prefix, prefix, contig->id);

    // Create contig specific files
    strcat(tcovFile, ".vertical");

    f = fOpen(tcovFile, "w");

    printInfo("Writing vertical coverage data for %s to %s\n", 
                 contig->id, tcovFile);
  }

  for (k = 0; k < MAXDEPTH; k++)
  {
    columnStatus[k] = 0; // Mark them all open
    columnID[k] = -1;
  }
  
  for (i = coords->start; i < coords->end; i++)
  {
    int k;
    int maxcolumn = -1;

    if (i > tilingEnd)
    {
      tilingStart = i;
      tilingEnd = loadTilingChunk(i, td, bcseq, contig, 0);
    }

    tilingpos = i - tilingStart;

    for (k = 0; k < MAXDEPTH; k++)
    {
      columnDisplay[k] = ' ';
    }

    for (j = 0; j < td->dcov[tilingpos]; j++)
    {
      int readid = td->idcov[tilingpos][j];
      int offset = bcseq[readid]->offset;
      char bc = bcseq[readid]->seq[(i-offset)%contig->len];

      int column = -1;

      // Try to find this read's column
      for (k = 0; k < MAXDEPTH; k++)
      {
        if (columnID[k] == readid)
        {
          column = k;
          break;
        }
      }

      if (column == -1)
      {
        for (k = 0; k < MAXDEPTH; k++)
        {
          if (columnStatus[k] == 0)
          {
            column = k;
            break;
          }
        }
      }

      if (column == -1)
      {
        printError("printVertical", DEBUG_ERROR,
                   "Can't find a column for %s at %d",
                   bcseq[readid]->id, i);
        exit(-1); // Can't find a column
      }

      columnStatus[column] = 3;
      columnDisplay[column] = bc;
      columnID[column] = readid;

      if (column > maxcolumn) { maxcolumn = column; }
    }


    //fprintf(f, "%d\t", i);

    for (k = 0; k <= maxcolumn; k++)
    {
      fprintf(f, "%c", columnDisplay[k]);

      if (columnStatus[k] > 0) { columnStatus[k]--; }
      if (columnStatus[k] == 0) { columnID[k] = -1; }
    }
    fprintf(f, "\n");
  }
  fprintf(f, "\n");

  if (f != stdout)
  {
    fclose(f);
  }
}

/*
  printHorizontal()
  
  cutAsm reborn
*/
void printHorizontal(struct CONTIG_SEQUENCE *contig,
                     struct CONTIG_SEQUENCE **bcseq,
                     tilingData *td, 
                     int showGapped,
                     Coordinates * coords,
                     const char * output_prefix)
{
  int i;
  int numOfSeqs = td->numOfSeqs;

  int maxval = coords->end;
  int numdigits = 1;
  int divide = 1;

  int * ungappedPos = NULL;

  unsigned int readid;

  // Assume stdout unless there is a prefix
  FILE * f = stdout;
  if (output_prefix && output_prefix != "-")
  {
    char buffer[FILENAME_MAX];
    strcpy(buffer, output_prefix);

    if (output_prefix[strlen(output_prefix)-1] == '/')
    {
      strcat(buffer, contig->id);
    }

    strcat(buffer, ".hcov");
    f = fOpen(buffer, "w");
    
    printInfo("Writing coverage data %s to %s\n", contig->id, buffer);
  }


  fprintf(f, "Assembly %s: ", contig->id);

  if (!showGapped)
  {
    int pos;
    int coordinate;
    int cur = coords->ugInd+coords->coordinateOffset;

    ungappedPos = newmem2(coords->end-coords->start, sizeof(int),
                          __FILE__, __LINE__);

    for (pos = coords->start; pos < coords->end; pos++)
    {
      if (contig->seq[pos] != '-')
      {
        coordinate = cur;
        cur++;
      }
      else
      {
        if (cur > coords->coordinateOffset) { coordinate = cur-1; }
        else                        { coordinate = cur; }
      }
      
      ungappedPos[pos-coords->start] = coordinate;
    }

    maxval = ungappedPos[coords->end-1-coords->start];
  }


  fprintf(f, "[%d%c, %d%c] (%d-based)\n\n",
          (showGapped) ? coords->start+coords->coordinateOffset : ungappedPos[0],
          (showGapped) ? 'G' : 'U',
          (showGapped) ? coords->end-1+coords->coordinateOffset : maxval,
          (showGapped) ? 'G' : 'U',
          coords->coordinateOffset);

  while (maxval >= 10)
  {
    divide *= 10;
    maxval /= 10;
    numdigits++;
  }

  for (; numdigits > 0; numdigits--, divide /= 10)
  {
//                                    Consensus:\t
    if (numdigits == 1) { fprintf(f, "Position: \t"); }
    else                { fprintf(f, "          \t"); }

    // Number the positions
    for (i=coords->start; i < coords->end; i++)
    {
      int curval = i+coords->coordinateOffset;

      if (!showGapped) { curval = ungappedPos[i-coords->start]; }

      curval = (curval/divide);
      //fprintf(f, "%c", (curval == 0) ? ' ' : (curval%10)+'0');
      fprintf(f, "%c", (curval%10)+'0');
    }

    fprintf(f, "\n");
  }
  fprintf(f, "\n");

  if (ungappedPos) { free(ungappedPos); }



  contig->seq[coords->end] = '\0';
  fprintf(f, "Consensus:\t%s\n\n", contig->seq+coords->start);
  fprintf(f, "Covering Reads:\n");


  for (readid = 0; readid < numOfSeqs; readid++)
  {
    int shift = 0;
    int offset = bcseq[readid]->offset;
    int len = bcseq[readid]->len;

    loadSeq(bcseq[readid]);

    fprintf(f, "%s\t", bcseq[readid]->id);
    
    if (offset >= 0)
    {
      if (offset <= coords->start)
      {
        shift = coords->start - offset;
      }
      else
      {
        for (shift = coords->start; shift < offset; shift++) 
          { fprintf(f, "."); }
        shift = 0;
      }

      if (offset + len > coords->end)
      {
        bcseq[readid]->seq[coords->end-offset] = '\0';
      }

      fprintf(f, "%s", bcseq[readid]->seq+shift);

      for (shift = offset+len; shift < coords->end; shift++) { fprintf(f, "."); }
    }
    else
    {
      int leftend = offset+len;
      int rightoffset = contig->len+offset;

      // take care of the left wrap-arround
      if (coords->start < leftend)
      {
        shift = coords->start - offset;

        if (leftend > coords->end)
        {
          bcseq[readid]->seq[coords->end-offset] = '\0';
        }

        fprintf(f, "%s", bcseq[readid]->seq+shift);
        shift = leftend;
      }
      else
      {
        shift = coords->start;
      }

      // print the interior spacer dots
      for (; shift < coords->end && shift < rightoffset; shift++)
      {
        fprintf(f, ".");
      }

      // now the right piece
      if (coords->end > rightoffset)
      {
        shift = 0;

        if (rightoffset < coords->start)
        {
          shift = coords->start - rightoffset;
        }

        bcseq[readid]->seq[coords->end - rightoffset] = '\0';
        fprintf(f, "%s", bcseq[readid]->seq+shift);
      }
    }

    fprintf(f, "\n");

    if (readid % 700 == 0) { pool_flush(); } // Should take most of a 1Mb block
  }

  pool_flush();

  fprintf(f, "\n");

  if (f != stdout) fclose(f);
}
