#include "ConsensusManager.hh"
#include <cstdio>

int ConsensusManager::m_doConsensusQC(0);
int ConsensusManager::m_doConsensusQV(0);
int ConsensusManager::m_doEditflag(0);
int ConsensusManager::m_defaultAmbiguityMode(3);
int ConsensusManager::m_maxPercDiscrepancy(33);

ConsensusManager::ConsensusManager() : RequestManager("ConsensusManager")
{
  m_log.debugL("Contructed ConsensusManager");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);

  m_includeSpace = false;

  registerOptionHandler("DoQualityClasses",   setDoConsensusQC);
  registerOptionHandler("DoQualityValue",     setDoConsensusQV);
  registerOptionHandler("DoEditflag",         setDoEditflag);
  registerOptionHandler("MaxPercDiscrepancy", setMaxPercDiscrepancy);
}

ConsensusManager::~ConsensusManager()
{
  m_log.debugL("Destructed ConsensusManager");
}

int ConsensusManager::getDefaultAmbiguityMode()
{
  return m_defaultAmbiguityMode;
}

void ConsensusManager::setDoConsensusQC(const string & value)
{
  m_doConsensusQC = (value == "0") ? 0 : 1;
}

void ConsensusManager::setMaxPercDiscrepancy(const string & value)
{
  m_maxPercDiscrepancy = atoi(value.c_str());
}

void ConsensusManager::setDoConsensusQV(const string & value)
{
  m_doConsensusQV = (value == "0") ? 0 : 1;
}

void ConsensusManager::setDoEditflag(const string & value)
{
  m_doEditflag = (value == "0") ? 0 : 1;
}

void ConsensusManager::addSlice(SliceData * slice)
{
  // Don't call RequestManager::addSlice because we need the results
  slice->setHighQualityThreshold(m_highQualityThreshold);
  slice->setScoringModel(m_scoringModel);
  slice->setDoAmbiguity(m_doAmbiguity); 

  if (m_doLogging)   { m_log << slice; }

  ConsensusData results = slice->doCalculations();

  if (m_doLogging)
  {
    m_log << slice; 
    m_log << "  <ConsensusDetails c=\"" << results.getConsensus()
          << "\" cbase=\""              << results.getConsensusBase()
          << "\" qc=\""                 << results.getQualityClass()
          << "\" qv=\""                 << results.getQualityValue()
          << "\" cp=\""                 << results.getProbability(' ')
          << "\" />" << endl; 

    char bases[] = {'A','C','G','T','-'};

    for (int i = 0; i < 5; i++)
    {
      m_log << "    <base=\"" << bases[i]
            << "\" add=\""    << results.testAmbiguityCode(bases[i])
            << "\" qv=\""     << results.getQualityValue(bases[i])
            << "\" cp=\""     << results.getProbability(bases[i])
            << "\" />" << endl; 
    }

    m_log << "  </ConsensusDetails>" << endl;
  }

  if (m_includeSpace)
  {
    if (m_doConsensusQC) { m_resultsQualityClass += " "; }
    if (m_doConsensusQV) { m_resultsQualityValue += " "; }
    if (m_doEditflag)    { m_resultsEditflag     += " ";  }
  }
  else
  {
    m_includeSpace = true;
  }

  m_resultsConsensus += results.getConsensus();

  int qc = results.getQualityClass();

  char buffer [16];


  if (m_doConsensusQC)
  {
    snprintf(buffer, sizeof(buffer), "%d", qc);
    m_resultsQualityClass += buffer;
  }

  if (m_doConsensusQV)
  {
    snprintf(buffer, sizeof(buffer), "%d", results.getQualityValue());
    m_resultsQualityValue += buffer;
  }

  if (m_doEditflag)
  {
    bool editflag = calculateEditflag(slice, qc, 
                                      results.getConsensusBase());

    m_resultsEditflag += (editflag) ? "1" : "0";
  }

  delete slice;
}

bool ConsensusManager::calculateEditflag(SliceData * slice, int qc, char cons)
{
  bool editFlag;

  editFlag = (qc >= 9);

  if (!editFlag)
  {
    int consNum;
    switch (cons)
    {
      case 'A': consNum = 0; break;
      case 'C': consNum = 1; break;
      case 'G': consNum = 2; break;
      case 'T': consNum = 3; break;
      case '-': consNum = 4; break;

      default: consNum = -1;
    };

    if (consNum != -1)
    {
      // The unambiguous consensus is well defined
      // Check if the threshold of discrepancy has been exceeded
      
      int againstCount = 0;
      int bccount [5] = {0,0,0,0,0}; // ACGT-

      const string & bases = slice->getNucData();
      int depth = slice->getDepth();

      for (int i = 0; i < depth; i++)
      {
        switch(bases[i])
        {
          case 'A':
          case 'a': bccount[0]++; break;

          case 'C':
          case 'c': bccount[1]++; break;

          case 'G':
          case 'g': bccount[2]++; break;

          case 'T':
          case 't': bccount[3]++; break;

          case '-': bccount[4]++; break;

          default:
            // Ambiguity code in sequence!
            againstCount++;
        };
      }


      for (int i = 0; i < 5; i++)
      {
        if (i != consNum)
        {
          againstCount += bccount[i];
        }
      }

      if ((depth > 3) && (againstCount * 100 >= m_maxPercDiscrepancy * depth))
      {
        // The threshold has been met
        editFlag = true;
      }
    }
    else
    {
      // Edit on unexpected input
      editFlag = true;
    }
  }

  return editFlag;
}

int ConsensusManager::printResults()
{
  if (!m_errorCount)
  {
    m_out << "<Consensus>" << endl;
    
    // Print the requested results

    if (m_doConsensus)
    {
      m_out << "  <Nuc>" 
            << m_resultsConsensus 
            << "</Nuc>" 
            << endl;
    }

    if (m_doConsensusQC)
    {
      m_out << "  <Qualclass>" 
            << m_resultsQualityClass 
            << "</Qualclass>" 
            << endl;
    }

    if (m_doConsensusQV)
    {
      m_out << "  <Qualval>" 
            << m_resultsQualityValue 
            << "</Qualval>" 
            << endl;
    }

    if (m_doEditflag)
    {
      m_out << "  <Editflag>" 
            << m_resultsEditflag 
            << "</Editflag>" 
            << endl;
    }

    m_out << "</Consensus>" << endl;
  }

  return m_errorCount;
}



void ConsensusManager::handleOptions(Foundation * tf)
{
  tf->getOptions()->addOptionResult("qualityvalue|v", &m_doConsensusQV,
                                    "Calculate the consensus quality value");

  tf->getOptions()->addOptionResult("qualityclass|q", &m_doConsensusQC,
                                    "Calculate the quality class");

  tf->getOptions()->addOptionResult("editflag|f", &m_doEditflag,
                                    "Calculate the editflag");

  RequestManager::handleOptions(tf);

  if (!m_doConsensus && !m_doConsensusQC && !m_doConsensusQV && !m_doEditflag)
  {
    //m_doConsensus = m_doConsensusQC = m_doEditflag = 1;
    m_doConsensus = m_doConsensusQC = 1;
  }

  if (!m_tf->getOptions()->hasOption("ambiguity"))
  {
    // Use Annotation as the default ambiguity method.
    m_doAmbiguity = m_defaultAmbiguityMode;
  }
}

