#include "SliceFattener.hh"
#include "ContigFattener.hh"

SliceFattener::SliceFattener() : RequestManager("SliceFattener")
{
  m_log.debugL("Contructed SliceFattener");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);

  m_contig = NULL;
}

int SliceFattener::s_displayType(0);

int SliceFattener::s_leftExtension(0);
int SliceFattener::s_rightExtension(0);
int SliceFattener::s_growContig(0);
int SliceFattener::s_fatten(0);

int SliceFattener::s_requireConfirmation(1);

int SliceFattener::s_fattenerRangeStart(-1);
int SliceFattener::s_fattenerRangeEnd(-1);
int SliceFattener::s_ungapped(0);

int SliceFattener::s_extensionDistance(0);
string SliceFattener::s_readToExtend("");

int SliceFattener::s_verbose(0);

SliceFattener::~SliceFattener()
{
  if (m_contig) delete m_contig;
  m_log.debugL("Destructed SliceFattener");
}

void SliceFattener::addSliceRange(long & index, long & gindex)
{
  m_contig = new Contig(index, gindex, m_readManager, s_verbose);

  index = gindex = 0;
}

void SliceFattener::addSlice(SliceData * slice)
{
  if (s_ungapped && s_fattenerRangeEnd != -1)
  {
    int ungapped = slice->getIndex();
    if (s_ungapped == 1 && ungapped >= s_fattenerRangeStart)
    {
      s_fattenerRangeStart = m_contig->getWidth();
      s_ungapped = 2;
    }
    else if (ungapped >= s_fattenerRangeEnd)
    {
      s_fattenerRangeEnd = m_contig->getWidth();
      s_ungapped = 0; // don't bother test any more
    }
  }

  m_contig->addSlice(slice);
}


void SliceFattener::addRequest(ConsensusRequest * request)
{
  RequestManager::addRequest(request);

  if (s_displayType == SLICE)
  {
    printXMLHeader(request);
  }
}


int SliceFattener::printResults()
{
  ContigFattener fattener(s_verbose);

  if (s_extensionDistance)
  {
    fattener.extendRead(m_contig, s_readToExtend, s_extensionDistance);
  }

  if (s_growContig)
  {
    if (s_rightExtension) { fattener.growContigRight(m_contig, s_requireConfirmation); }
    if (s_leftExtension)  { fattener.growContigLeft(m_contig,  s_requireConfirmation); }
  }

  if (s_fatten)
  {
    fattener.fattenContig(m_contig,
                          s_fattenerRangeStart, 
                          s_fattenerRangeEnd,
                          s_leftExtension,
                          s_rightExtension);
  }

//  ContigOptimizer optimizer;
//  optimizer.optimize(m_contig);

  m_contig->printContig(m_out, (PrintType) s_displayType);

  if (s_displayType == SLICE) { m_out << "</Request>" << endl; }

  return m_errorCount;
}


void SliceFattener::handleOptions(Foundation * tf)
{
  int unconfirmed = !s_requireConfirmation;

  tf->getOptions()->addOptionResult("x=i", &s_fattenerRangeStart,
                                    "Specify start of range to fatten");

  tf->getOptions()->addOptionResult("y=i", &s_fattenerRangeEnd,
                                    "Specify start of range to fatten");

  tf->getOptions()->addOptionResult("d=i", &s_extensionDistance,
                                    "Specify distance to extend");

  tf->getOptions()->addOptionResult("r=s", &s_readToExtend,
                                    "Specify read to extend");

  tf->getOptions()->addOptionResult("R", &s_rightExtension,
                                    "Perform Right Extension while fattening or growing");

  tf->getOptions()->addOptionResult("L", &s_leftExtension,
                                    "Perform Left Extension while fattening or growing");

  tf->getOptions()->addOptionResult("T=i", &s_displayType,
                                    "Display Type (0=slice, 1=tcov, 2=contig)");

  tf->getOptions()->addOptionResult("g|grow", &s_growContig,
                                    "Grow the contig");

  tf->getOptions()->addOptionResult("f|fatten", &s_fatten,
                                    "Fatten the contig in the optional range");

  tf->getOptions()->addOptionResult("v|verbose", &s_verbose,
                                    "Be verbose when reporting");

  tf->getOptions()->addOptionResult("U|ungapped", &s_ungapped,
                                    "Ranges are ungapped coordinates");

  tf->getOptions()->addOptionResult("unconfirmed", &unconfirmed,
                                    "Do unconfirmed contig growing");

  RequestManager::handleOptions(tf);

  s_requireConfirmation = !unconfirmed;

  // Sanity check that a read and a distance are specified together
  if ((!s_readToExtend.empty() && !s_extensionDistance) ||
      ( s_readToExtend.empty() &&  s_extensionDistance))
  {
    throw Exception("You must specify a readname and a distance to extend together",
                        "SliceFattener");
  }

  // Enable both by default
  if (!s_leftExtension && !s_rightExtension)
  {
    s_leftExtension = s_rightExtension = 1;
  }

  // By default, just fatten
  if (!s_fatten && !s_growContig && !s_extensionDistance)
  {
    s_fatten = 1;
  }
}

