#include "ContigSequence.hh"
#include "RequestManager.hh"

ContigSequence::ContigSequence(ReadData * rd, long gindex, int readCount)
{
  m_offset = gindex;
  m_readData = rd;
  m_readCount = readCount;
  m_length = 0;
}

ContigSequence::~ContigSequence()
{

}

int ContigSequence::getReadCount() const
{
  return m_readCount;
}

const string & ContigSequence::getSeqname() const
{
  return m_readData->getSeqname();
}

bool ContigSequence::isReverseCompliment() const
{
  return m_readData->isReverseCompliment();
}

void ContigSequence::printReadData(ostream & out)
{
  m_readData->printXML(out);
}

void ContigSequence::printContigSequence(ostream & out, long asm_l, long asm_r)
{
  out << "#" << m_readData->getSeqname()
      << "(" << m_offset << ")";
  
  if (m_readData->isReverseCompliment()) { out << " [RC] "; }
  else { out << " [] "; }

  out << m_nuc.length() << " bases, 00000000 checksum. "
      << "{"   << m_readData->getSeqleft() 
      << " "   << m_readData->getSeqright()
      << "} <" << asm_l
      << " "   << asm_r
      << ">" << endl;

  RequestManager::print60(out, m_nuc);
}

void ContigSequence::addSlice(char base, char qv)
{
  m_nuc.push_back(base);
  m_qual.push_back(qv);
  m_length++;
}

void ContigSequence::addSequence(const string & bases,
                                 const string & quals)
{
  m_nuc += bases;
  m_length += bases.size();

  //TODO: parse qual string
}

const char * ContigSequence::getNuc(long gindex)
{
  long position = gindex - m_offset;

  if (position < 0 || position >= m_length)
  {
    cerr << "Nuc requested outside of range in r" << m_readCount
         << " gindex: " << gindex 
         << " offset: " << m_offset
         << " length: " << m_length
         << endl;

    throw Exception("Nuc requested outside of range!",
                    "ContigSequence::getNuc");
  }

  char base = m_nuc[position];

  static char retval [2];
  retval[0] = base;
  retval[1] = '\0';

  return retval;
}

const char * ContigSequence::getQual(long gindex)
{
  long position = gindex - m_offset;
  if (position < 0 || position >= m_length)
  {
    cerr << "Qual requested outside of range in r" << m_readCount
         << " gindex: " << gindex 
         << " offset: " << m_offset
         << " length: " << m_length
         << endl;

    throw Exception("Qual requested outside of range!",
                    "ContigSequence::getQual");
  }

  char qv = m_qual[position];

  static char retval [16];

  snprintf(retval, sizeof(retval), "%d", qv);

  return retval;
}

const char * ContigSequence::getReadId()
{
  return m_readData->getId().c_str();
}

char & ContigSequence::operator[](long gindex)
{
  long position = gindex - m_offset;

  return m_nuc[position];
}

char ContigSequence::base(long gindex)
{
  if (gindex < m_offset || gindex >= m_offset + m_length)
  {
    cerr << "Invalid position requested: "<< gindex 
         << " in r" << m_readCount << endl;
    throw Exception("Invalid position requested", "ContigSequence::base");
  }

  long position = gindex - m_offset;
  return m_nuc[position];
}

int ContigSequence::qv(long gindex) const
{
  if (gindex < m_offset || gindex >= m_offset + m_length)
  {
    cerr << "Invalid position requested: "<< gindex 
         << " in r" << m_readCount << endl;
    throw Exception("Invalid position requested", "ContigSequence::qv");
  }

  long position = gindex - m_offset;
  return m_qual[position];
}

float ContigSequence::getAvgQV3Trim(int count) const
{
  const string & qualstring = m_readData->getRawQualData(false);
  int i;

  int qualsum = 0;
  int qualval;
  string::size_type current = 0;

  for (i = 0; current < qualstring.length() && i < count; i++)
  {
    string::size_type endpos = qualstring.find(' ', current);
    const string & qualstr = qualstring.substr(current, endpos-current);
    qualval = atoi(qualstr.c_str());

    qualsum += qualval;
    current = endpos+1;
  }

  return (i) ? ((float)qualsum)/i : 0.0;
}

float ContigSequence::getAvgQV(long offset, int count) const
{
  int qvsum = 0;
  int i;

  if (isReverseCompliment())
  {
    for (i = 0; i < count && offset+i < m_offset + m_length; i++) 
      { qvsum += qv(offset+i); }
  }
  else
  {
    for (i = 0; i < count && offset-i > m_offset; i++) 
      { qvsum += qv(offset-i); }
  }

  return (i) ? ((float) qvsum) / i : 0.0;
}


void ContigSequence::adjustOffset(long gindex, int delta)
{
  if (m_offset >= gindex)
  {
    m_offset += delta;
  }
}

void ContigSequence::swap(long a, long b)
{
  char baseswap = m_nuc[a];
  m_nuc[a] = m_nuc[b];
  m_nuc[b] = baseswap;

  if (!m_qual.empty())
  {
    char qualswap = m_qual[a];
    m_qual[a] = m_qual[b];
    m_qual[b] = qualswap;
  }
}

int ContigSequence::leftTwiddle(long gindex, int & twiddleDistance, char cons)
{
  int retval = 0;
  long position = gindex - m_offset;
  if (position > 0 && position < m_length)
  {
    if (twiddleDistance == 0)
    {
      char base;
      twiddleDistance = 1;
      while (position-twiddleDistance >= 0 &&
             ((base = m_nuc[position-twiddleDistance]) == '-'))
      {
        twiddleDistance++;
      }

      if (position-twiddleDistance >= 0)
      {
        if (cons != '?' && base != cons)
        {
          twiddleDistance = 0;
        }
      }
    }

    if (twiddleDistance && position - twiddleDistance >= 0)
    {
      swap(position-twiddleDistance, position);
      retval = 1;
    }
  }

  return retval;
}

int ContigSequence::rightTwiddle(long gindex, int & twiddleDistance, char cons)
{
  int retval = 0;
  long position = gindex - m_offset;

  if ((position >= 0) && (position+1 < m_length))
  {
    if (twiddleDistance == 0)
    {
      char base;
      twiddleDistance = 1;
      while (position+twiddleDistance < m_length &&
             ((base = m_nuc[position+twiddleDistance]) == '-'))
      {
        twiddleDistance++;
      }

      if (position+twiddleDistance < m_length)
      {
        if (cons != '?' && base != cons)
        {
          twiddleDistance = 0;
        }
      }
    }

    if (twiddleDistance && position + twiddleDistance < m_length)
    {
      swap(position, position+twiddleDistance);
      retval = 1;
    }
  }

  return retval;
}

void ContigSequence::erase(long gindex)
{
  long position = gindex - m_offset;

  if (position >= 0 && position < m_length)
  {
    if (m_nuc[position] != '-')
    {
      cerr << "Error: erasing " << m_nuc[position] << " at " << position 
           << " in r" << m_readCount << endl;

      throw Exception("erasing non-gap element from read!",
                      "ContigSequence::erase");
    }

    m_nuc.erase(m_nuc.begin() + position);
    if (!m_qual.empty()) { m_qual.erase(m_qual.begin() + position); }

    m_length--;
  }
  else if (position < 0)
  {
    // A slice was erased to the left of this read
    m_offset--;
  }
}

void ContigSequence::dumpRange(long gindexstart, long gindexend)
{
  if (!(m_offset > gindexend || m_offset+m_length < gindexstart))
  {
    cerr << "r[" << m_readCount << "]:\t";

    for (int i = gindexstart; i < m_offset; i++)
    {
      cerr << '*';
    }

    long startpos = gindexstart - m_offset;
    long endpos = gindexend - m_offset;

    if (startpos < 0)       { startpos = 0; }
    if (endpos >= m_length) { endpos = m_length-1; }

    cerr << m_nuc.substr(startpos, endpos-startpos+1) << endl;
  }
}

void ContigSequence::dump()
{
  cerr << "r[" << m_readCount << "]:\t";
  for (int i = 0; i< m_offset; i++)
  {
    cerr << '*';
  }

  cerr << m_nuc << endl;
}

long ContigSequence::getRightEnd() const
{
  // minus 1 or we double count the offset
  return m_offset + m_length - 1;
}

long ContigSequence::getOffset() const
{
  return m_offset;
}

long ContigSequence::stripRightGaps()
{
  int rightGaps = 0;
  while (m_nuc[m_length-1-rightGaps] == '-')
  {
    rightGaps++;
  }

  if (rightGaps)
  {
    cerr << "Stripping " << rightGaps << 
            " rightGaps from r" << m_readCount << endl;

    m_nuc.resize(m_nuc.size()-rightGaps);
    if (!m_qual.empty()) { m_qual.resize(m_qual.size()-rightGaps); }
    
    m_length -= rightGaps;
  }

  return getRightEnd();
}

long ContigSequence::stripLeftGaps()
{
  int leftGaps = 0;

  while (m_nuc[leftGaps] == '-')
  {
    leftGaps++;
  }

  if (leftGaps)
  {
    cerr << "Stripping " << leftGaps << 
            " leftGaps from r" << m_readCount << endl;

    m_nuc.erase(0, leftGaps);
    if (!m_qual.empty()) { m_qual.erase(m_qual.begin(), m_qual.begin()+leftGaps); }
    
    m_offset += leftGaps;
    m_length -= leftGaps;
  }

  return m_offset;
}

int ContigSequence::get3TrimLength()
{
  // false always means 3' trim
  return m_readData->getRawNucData(false).length();
}

string ContigSequence::getLeftTrim()
{
  string reversed;

  // We are getting the right trim, but reverse complimenting it
  m_leftTrim = m_readData->getRawNucData(false);
  m_leftTrimQual = m_readData->getRawQualData(false);

  m_leftTrimReversed = "";
  for (int i = m_leftTrim.size()-1; i >= 0; i--)
  {
    char base = m_leftTrim[i];
    m_leftTrimReversed.push_back(libSlice_getCompliment(base));
  }

  return m_leftTrimReversed;
}

string ContigSequence::getRightTrim()
{
  m_rightTrim = m_readData->getRawNucData(false);
  m_rightTrimQual = m_readData->getRawQualData(false);
  return m_rightTrim;
}

int ContigSequence::extractFirstQuality(string & qualstring, bool doErase)
{
  int qualval = 0;

  string::size_type endpos = qualstring.find(' ');

  if (endpos == qualstring.npos)
  {
    qualval = atoi(qualstring.c_str());
    if (doErase) { qualstring = ""; }
  }
  else
  {
    const string & qualstr = qualstring.substr(0, endpos);
    qualval = atoi(qualstr.c_str());
    if (doErase) { qualstring.erase(0, endpos+1); }
  }

  return qualval;
}

int ContigSequence::extendLeft(const string & extendLeft)
{
  int basesExtended = 0;

  if (!extendLeft.empty())
  {
    //cerr << "Left Extended r" << m_readCount
    //     << ": " << extendLeft << endl;
    
    unsigned int addedBases = extendLeft.size();
    m_length += addedBases;
    m_offset -= addedBases;

    m_nuc = extendLeft + m_nuc;
    int qualval = m_qual[0];

    for (int i = addedBases-1; i >= 0; i--)
    {
      if (extendLeft[i] != '-')
      {
        basesExtended++;
        qualval = extractFirstQuality(m_leftTrimQual, true);
        m_leftTrim.erase(0, 1);
      }
      else
      {
        if (!m_leftTrimQual.empty())
        {
          int nextqv = extractFirstQuality(m_leftTrimQual, false);
          if (nextqv < qualval) { qualval = nextqv; }
        }
      }

      m_qual.insert(m_qual.begin(), qualval);
    }

    m_readData->flushTrimmedSeqSide(false);

    m_readData->addNucData(m_leftTrim.c_str(), false);
    m_readData->addQualData(m_leftTrimQual.c_str(), false);

    long clrleft = m_readData->getClrleft();
    long clrright = m_readData->getClrright();

    m_readData->setClrData(clrleft, clrright + basesExtended);

    /*
    cerr << "Left Extension of r" << m_readCount 
         << " by " << basesExtended 
         << " (" << addedBases << ")" << endl;
    */
  }

  return basesExtended;
}

int ContigSequence::extendRight(const string & extendRight)
{
  int basesExtended = 0;

  if (!extendRight.empty())
  {
    // cerr << "Extended sequence: " << extendRight 
    //     << " in r" << m_readCount << endl;
    
    unsigned int addedBases = extendRight.length();
    m_length += addedBases;

    m_nuc += extendRight;
    int qualval = m_qual[m_qual.size()-1];

    for (unsigned int i = 0; i < addedBases; i++)
    {
      if (extendRight[i] != '-')
      {
        basesExtended++;
        qualval = extractFirstQuality(m_rightTrimQual, true);
        m_rightTrim.erase(0, 1);
      }
      else
      {
        if (!m_rightTrimQual.empty())
        {
          int nextqv = extractFirstQuality(m_rightTrimQual, false);
          if (nextqv < qualval) { qualval = nextqv; }
        }
      }

      m_qual.push_back(qualval);
    }

    m_readData->flushTrimmedSeqSide(false);

    m_readData->addNucData(m_rightTrim.c_str(), false);
    m_readData->addQualData(m_rightTrimQual.c_str(), false);

    long clrleft = m_readData->getClrleft();
    long clrright = m_readData->getClrright();

    m_readData->setClrData(clrleft, clrright + basesExtended);

    /*
    cerr << "Right Extension of r" << m_readCount 
         << " by " << basesExtended 
         << " (" << addedBases << ")" << endl;
    */
  }

  return basesExtended;
}

void ContigSequence::insertGap(long gindex)
{
  long position = gindex - m_offset;

  /*
  cerr << "Inserting gap into r" << m_readCount 
       << " at " << position << endl;
  */

  if (position <= 0) // don't insert gap as first character
  {
    m_offset++;
  }
  else if (position < m_length) // don't insert gap as last character
  {
    m_nuc.insert(position, "-");

    if (!m_qual.empty()) 
    { 
      // We never insert gap as first or last base, so there should
      // always be two flanking qvs, but gracefully handle the case
      // were there is not
      
      int leftqv  = (position) ? m_qual[position-1] : -1;
      int rightqv = (position < m_qual.size()) ? (m_qual[position]) : -1;

      // if (leftqv == -1 || rightqv == -1) { cerr << "*** Unexpected qv l: " << leftqv << " r: " << rightqv << endl; }

      int minqv = (leftqv < rightqv) 
                ? (leftqv == -1)  ? rightqv : leftqv
                : (rightqv == -1) ? leftqv  : rightqv;

      m_qual.insert(m_qual.begin()+position, minqv); 
    }

    m_length++;
  }
}



// gindex is the new start of the read
int ContigSequence::trimReadLeft(long gindex)
{
  int difference = 0;

  if (gindex >= m_offset)
  {
    //cerr << "Old coordinates: " << m_offset << "," << getRightEnd() << endl;

    string trimBases;
    string trimQual;
    char buffer[16];

    for (int i = m_offset; i < gindex; i++)
    {
      int seqpos = i - m_offset;
      
      //reverse compliment the bases to remove
      char base = libSlice_getCompliment(m_nuc[seqpos]);
      trimBases.insert(trimBases.begin(), base);

      sprintf(buffer, "%02d ", m_qual[seqpos]);
      trimQual.insert(0, buffer);
    }

    trimQual.erase(trimQual.begin()+trimQual.size()-1);
    difference = trimBases.size();

    //cerr << "Found " << difference << " bp to trim" << endl;

    m_offset = gindex;
    m_nuc.erase(0, difference);
    m_qual.erase(m_qual.begin(), m_qual.begin()+difference);

    m_length = m_nuc.size();

    m_readData->flushTrimmedSeqSide(false);

    m_leftTrim.insert(0, trimBases); 
    m_leftTrimQual.insert(0, trimQual);

    m_readData->addNucData(m_leftTrim.c_str(), false);
    m_readData->addQualData(m_leftTrimQual.c_str(), false);

    long clrleft = m_readData->getClrleft();
    long clrright = m_readData->getClrright();

    m_readData->setClrData(clrleft, clrright - difference);

    //cerr << "New coordinates: " << m_offset << "," << getRightEnd() << endl;
  }

  return difference;
}

int ContigSequence::trimReadRight(long gindex)
{
  long oldend = m_offset + m_length - 1;
  int difference = 0;

  if (oldend >= gindex)
  {
    //cerr << "Old coordinates: " << m_offset << "," << getRightEnd() << endl;
    m_length = gindex - m_offset;
    difference = oldend - gindex + 1;
    //cerr << "Difference: " << difference << endl;
    //cerr << "New coordinates: " << m_offset << "," << getRightEnd() << endl;
    
    // push the bases and quals back out of the clear range
    
    string trimBases;
    string trimQual;
    char buffer[16]; 
    
    for (int i = m_length; i < m_nuc.size(); i++)
    {
      trimBases += m_nuc[i];
      sprintf(buffer, "%02d", m_qual[i]);
      trimQual += buffer;
      trimQual += " ";
    }

    if (m_rightTrim.empty())
    {
      // erase the trailing space
      trimQual.erase(trimQual.size()-1,1);
    }

    m_nuc.resize(m_length);
    m_qual.resize(m_length);

    m_rightTrim += trimBases;
    m_rightTrimQual += trimQual;

    m_readData->flushTrimmedSeqSide(false);
    m_readData->addNucData(m_rightTrim.c_str(), false);
    m_readData->addQualData(m_rightTrimQual.c_str(), false);

    long clrleft = m_readData->getClrleft();
    long clrright = m_readData->getClrright();

    m_readData->setClrData(clrleft, clrright - trimBases.size());
  }

  return difference;
}

