#include "ConsensusSequence.hh"
#include <stdio.h> // snprintf
#include <time.h>  // time for ALPHA

using libSlice::ConsensusData;

ConsensusSequence::ConsensusSequence(int contigCount, 
                                     long index, 
                                     ostream & out)
  : m_out(out),
    m_nuc(""),
    m_nucsequence(""),
    m_qualclass("0x")
{
  m_contigCount = contigCount;

  m_curOffset = 0;
  m_curIndex  = 1;

  m_depthCount = 0;
  m_sliceCount = 0;
  m_ambiguityCount = 0;
}

ConsensusSequence::~ConsensusSequence()
{
  SequenceList::const_iterator i;
  for (i =  m_sequenceList.begin();
       i != m_sequenceList.end();
       i++)
  {
    delete *i;
  }
}

void ConsensusSequence::addSlice(SliceData * slice)
{
  char consensus = slice->getConsensus();
  ConsensusData result = slice->doCalculations();

  m_nuc.push_back(consensus);

  long coordinate = m_curIndex;
  if (consensus == '-') { coordinate--; }
  else                  { m_nucsequence.push_back(consensus); }

  char buffer[8];
  snprintf(buffer, sizeof(buffer), "%02X", result.getQualityClass());
  m_qualclass += buffer;

  unsigned short depth = slice->getDepth();
  for (unsigned short i = 0; i < depth; i++)
  {
    SequenceData * sd = getSequence(slice->getRead(i), m_curOffset, coordinate);

    sd->addElement(coordinate, slice->getBase(i));
  }

  m_curOffset++;
  if (consensus != '-') { m_curIndex++; }

  if (depth)
  {
    m_sliceCount++;
    m_depthCount += depth;
    m_ambiguityCount += 1 - (consensus == 'A' || consensus == 'C' ||
                             consensus == 'T' || consensus == 'G' ||
                             consensus == '-');
  }
}


SequenceData * ConsensusSequence::getSequence(const ReadData * rd, 
                                              long offset,
                                              long index)
{
  SequenceData * retval;
  SequenceMap::const_iterator i;

  i = m_sequenceMap.find(rd->getId());

  if (i == m_sequenceMap.end())
  {
    retval = new SequenceData(rd, offset, index);
    m_sequenceMap[rd->getId()] = retval;
    m_sequenceList.push_back(retval);
  }
  else
  {
    retval = i->second;
  }

  return retval;
}


void ConsensusSequence::printTasm(const string & asmblId,
                                  const string & user,
                                  const string & comment,
                                  const string & method,
                                  const string & caId,
                                  string & mod_date,
                                  int doQuality)
{ 
  string date = mod_date;

  if (date == "")
  {
    time_t now = time(NULL);
    char timeStr [128];
    strftime(timeStr, sizeof(timeStr), "%D %r", localtime(&now));

    date = timeStr;
  }

  char redundancy[16];
  snprintf(redundancy, sizeof(redundancy), "%.2f", 
          ((float)m_depthCount)/m_sliceCount);

  char perc_N[16];
  snprintf(perc_N, sizeof(perc_N), "%.2f", 
           ((float)m_ambiguityCount)/m_sliceCount);

  m_out << "sequence\t"     << m_nucsequence          << endl
        << "lsequence\t"    << m_nuc                  << endl;

  if (doQuality)
  {
    m_out << "quality\t"      << m_qualclass            << endl;
  }

  m_out << "asmbl_id\t"     << asmblId                << endl
        << "ca_contig_id\t" << caId                   << endl
        << "seq_id\t"                                 << endl
        << "com_name\t"                               << endl
        << "type\t"                                   << endl
        << "method\t"       << method                 << endl
        << "ed_status\t"                              << endl
        << "redundancy\t"   << redundancy             << endl
        << "perc_N\t"       << perc_N                 << endl
        << "seq#\t"         << m_sequenceList.size()  << endl
        << "full_cds\t"                               << endl
        << "cds_start\t"                              << endl
        << "cds_end\t"                                << endl
        << "ed_pn\t"        << user                   << endl
        << "ed_date\t"      << date                   << endl
        << "mod_date\t"     << date                   << endl
        << "comment\t"      << comment                << endl
        << "frameshift\t"                             << endl;

  SequenceList::const_iterator i;
  for (i =  m_sequenceList.begin();
       i != m_sequenceList.end();
       i++)
  {
    m_out << endl;
    (*i)->printTasmRecord(m_out);
  }
}
