#include "SliceZipper.hh"
#include <cstdio>

SliceZipper::SliceZipper()
  : RequestManager("SliceZipper"),
    m_readMerger(m_tf, m_out, m_uberReadManager)
{
  m_rangeCount = 0;
  m_currentReferencePosition = 0;
  m_currentReferenceUngapped = 0;
  m_currentQueryPosition = 0;
  m_currentQueryUngapped = 0;
  m_mergedSlices = false;

  m_testReferenceGaps = false;

  m_newIndex = 0;
  m_newGindex = 0;

  m_lastslice = NULL;

  m_log.debugL("Contructed SliceZipper");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_INFORMATION);

  if (!m_queryGaps.empty())
  {
    m_log << "Loading Query Gaps" 
          << (m_reverseQuery ? "(reversed)" : "") << endl;
    loadGaps(m_queryGaps, m_queryGapMap);
  }

  if (!m_referenceGaps.empty())
  {
    m_log << "Loading Reference Gaps" << endl;
    loadGaps(m_referenceGaps, m_referenceGapMap);
  }
}

int SliceZipper::m_alignStart(0);
int SliceZipper::m_promoteAlignmentGaps(1);
int SliceZipper::m_reverseQuery(0);
int SliceZipper::m_ungappedCoordinates(0);
int SliceZipper::m_gapCoordinateOffset(0);
int SliceZipper::m_referenceConsensus(0);

string SliceZipper::m_queryGaps;
string SliceZipper::m_referenceGaps;

SliceZipper::~SliceZipper()
{
  m_log.debugL("Destructed SliceZipper");
  if (m_lastslice) { delete m_lastslice; }
}

void SliceZipper::addRequest(ConsensusRequest * request)
{
  if (!m_requestData)
  {
    RequestManager::addRequest(request);

    printXMLHeader(request);

    m_out << "  <ReadCollection>" << endl;
  }
  else
  {
    delete request;
  }
}

void SliceZipper::addReadData(ReadData * readData)
{
  if (m_reverseQuery && m_rangeCount == 0)
  {
    readData->reverseCompliment();
  }

  m_readMerger.addReadData(readData);
}


void SliceZipper::addSliceRange(long & index, long & gindex)
{
  m_rangeCount++;

  index = gindex = 0L;

  m_origRangeIndex = 0;
  m_origRangeGindex = 0;

  m_currentQueryPosition = 0;
  m_currentQueryUngapped = 0;
  m_currentReferencePosition = 0;
  m_currentReferenceUngapped = 0;

  if (m_rangeCount == 2)
  {
    m_out << "  </ReadCollection>"    << endl
          << "  <SliceRange Index=\"" << m_origRangeIndex
          << "\" Gindex=\""           << m_origRangeGindex
          << "\">" << endl;

    VectorSlice * querySlice = NULL;

    // Test if query slices left flank merge region
    if (m_alignStart > 0)
    {
      while (m_alignStart > 0)
      {
        if (m_sliceResults.empty())
        {
          m_log.fatal("Invalid alignment offset! Type 1");
          throw Exception("Invalid alignment offset!", "SliceZipper");
        }

        querySlice = *m_sliceResults.begin();
        m_sliceResults.pop_front();

        if (!m_ungappedCoordinates || (querySlice->getConsensus() != '-'))
        {
          m_alignStart--;
        }

        if (!querySlice->isIntraGap()) 
        { 
          m_currentQueryPosition++; 
          if (querySlice->getConsensus() != '-') { m_currentQueryUngapped++; }
        }

        printSlice(querySlice);
      }

      if (m_sliceResults.empty())
      {
        m_log.fatal("Invalid alignment offset! Type 2");
        throw Exception("Invalid alignment offset!", "SliceZipper");
      }
      
      // Special case where there are gaps in the query where the alignment
      // would otherwise begin
      if (m_ungappedCoordinates)
      {
        querySlice = *m_sliceResults.begin();

        while (querySlice &&
               querySlice->getConsensus() == '-')
        {
          printSlice(querySlice);
          m_sliceResults.pop_front();

          if (m_sliceResults.empty())
          {
            m_log.fatal("Invalid alignment offset! Type 3");
            throw Exception("Invalid alignment offset!", "SliceZipper");
          }

          // Don't advance ungapped
          if (!querySlice->isIntraGap()) { m_currentQueryPosition++; }
          querySlice = *m_sliceResults.begin();
        }
      }
    }
  }
}

void SliceZipper::mergeSlice(VectorSlice * slice, 
                             const string & nucData,
                             const string & qualData,
                             const string & readData,
                             char refcons,
                             bool isGap)
{
  if (m_referenceConsensus)
  {
    slice->setConsensus(refcons);
  }

  // Leave the original slice->consensus value
  if (slice->getNucData() != "" && !nucData.empty())
  {
    // Add spaces as necessary (no space between nucs)
    if (qualData[0] != ' ') { slice->addQualVal(" "); }
    if (readData[0] != ' ') { slice->addReadId(" ");  }
  }

  // Call reset so that the flattenedSlice will be updated
  string origNuc(slice->getNucData());
  slice->resetNuc();
  slice->addNuc(origNuc.c_str());

  slice->addNuc(nucData.c_str());
  slice->addQualVal(qualData.c_str());
  slice->addReadId(readData.c_str());

  if (isGap)
  {
    slice->setGap(true);
  }
}

VectorSlice * SliceZipper::createVectorSlice(bool intraGap,
                                             bool isGap,
                                             const string & nucData,
                                             const string & qualData,
                                             const string & readData,
                                             const char * consensus)
{
  // Just create a new slice. This has the advantage of being tied to
  // the uberReadManager, and the memory can be safely deleted from the
  // original
  VectorSlice * newSlice = new VectorSlice(intraGap, isGap, m_uberReadManager, 
                                           "0", "0", consensus, 0, 0);
  newSlice->addNuc(nucData.c_str());
  newSlice->addQualVal(qualData.c_str());
  newSlice->addReadId(readData.c_str());

  return newSlice;
}

void SliceZipper::createGapValues(SliceData * slice,
                                  string & nuc,
                                  string & qual)
{
  int sdepth = slice->getDepth();
  char buffer [16];

  // Create the correct number of gap elements for the new slice
  for(int i = 0; i < sdepth; i++)
  {
    if (i) qual += " ";

    sprintf(buffer, "%d", slice->getQualVal(i));
    qual += buffer;
    nuc += "-";
  }
}

void SliceZipper::printSlice(VectorSlice * slice)
{
  if (m_lastslice && slice->isGap())
  {
    int depth = slice->getDepth();
    int odepth = m_lastslice->getDepth();

    string qualstr;
    char buffer[16];
    int change = 0;

    // Gap quality values have been initialized from the right neighboring
    // slice. m_lastslice stores the left slice, so it is a simple matter
    // to compute the minimum

    for (int i = 0; i < depth; i++)
    {
      if (i) { qualstr += " "; }
      int qv = ((SliceData *)slice)->getQualVal(i);
      const string & id = slice->getRead(i)->getId();
      int found = 0;

      if (slice->getBase(i) == '-')
      {
        for (int j = 0; j < odepth; j++)
        {
          if (id == m_lastslice->getRead(j)->getId())
          {
            int lastqual = m_lastslice->getQualVal(j);

            if (lastqual < qv)
            {
              qv = lastqual;
              change = 1;
              found = 1;
              break;
            }
          }
        }

        if (!found)
        {
          // TODO: Strip this edge gap, but be careful in case of circular zip
        }
      }

      sprintf(buffer, "%d", qv); 
      qualstr += buffer;
    }

    if (change)
    {
      slice->resetQualVal();
      slice->addQualVal(qualstr.c_str());
    }
  }

  RequestManager::addSlice(slice);
  char consensus = slice->getConsensus();

  long coordinate = m_newIndex;
  if (coordinate && consensus == '-') coordinate--;
  slice->resetIndex(coordinate, m_newGindex);
  slice->printXML(m_out);

  delete m_lastslice;
  m_lastslice = slice;

  if (consensus != '-') m_newIndex++;
  m_newGindex++;
}

int SliceZipper::getGapCount(bool isReference, int gpos, int pos)
{
  int retval = 0;
  GapMap::iterator i;
  
  int location = (m_ungappedCoordinates) ? pos : gpos;

  if (isReference)
  {
    i = m_referenceGapMap.find(location);

    if (i != m_referenceGapMap.end())
    {
      retval = i->second;
      i->second = 0;
    }
  }
  else
  {
    i = m_queryGapMap.find(location);

    if (i != m_queryGapMap.end())
    {
      retval = i->second;
      i->second = 0;
    }
  }

  return retval;
}

void SliceZipper::storeQuerySlice(VectorSlice * query)
{
  if (m_reverseQuery)
  {
    // store in reverse order (in a stack)
    m_sliceResults.push_front(query);
  }
  else
  {
    // store in the vector in the order the slices came in
    m_sliceResults.push_back(query);
  }
}

void SliceZipper::handleQuerySlice(SliceData * query, 
                                   const string & newids,
                                   const char * consensus)
{
  VectorSlice * newSlice = NULL;

  if (m_reverseQuery)
  {
    char newconsensus[2];
    newconsensus[0] = libSlice_getCompliment(consensus[0]);
    newconsensus[1] = '\0';

    string reversednucs;
    unsigned int depth = query->getDepth();
    for (unsigned int i=0; i < depth; i++)
    {
      reversednucs += libSlice_getCompliment(query->getBase(i));
    }

    newSlice = createVectorSlice(false,
                                 false,
                                 reversednucs,
                                 query->getQualValData(),
                                 newids,
                                 newconsensus);
  }
  else
  {
    // Create a new vector slice to hold the final slice
    newSlice = createVectorSlice(false,
                                 false,
                                 query->getNucData(),
                                 query->getQualValData(),
                                 newids,
                                 consensus);
  }

  if (m_currentQueryPosition == 0)
  {
    // There may be leading gaps (at -1) to promote
    int gapsToAdd = getGapCount(false, -1, -1);
    
    while (gapsToAdd)
    {
      m_log << "Promoting gap slice into query at -1" << endl;

      string nuc, qual;
      createGapValues(query, nuc, qual);

      storeQuerySlice(createVectorSlice(true, true, nuc, qual, newids, "-"));

      gapsToAdd--;
    }
  }
  
  storeQuerySlice(newSlice);

  // The compliment of - is -, so it's safe to test the original consensus
  int coordinate = m_currentQueryUngapped;
  if (coordinate && consensus[0] == '-') { coordinate--; }

  // Store the query intra gaps in the vector
  // It's possible to add meaningless intra-gaps, since we don't know if
  // we are in a region that aligns, but it doesn't hurt.
  int gapsToAdd = getGapCount(false, 
                              m_currentQueryPosition, 
                              coordinate);
  
  m_currentQueryPosition++;
  if (consensus[0] != '-') { m_currentQueryUngapped++; }

  while (gapsToAdd)
  {
    m_log << "Promoting gap slice into query at " 
          << m_currentQueryPosition << endl;

    string nuc, qual;
    createGapValues(query, nuc, qual);

    storeQuerySlice(createVectorSlice(true, true, nuc, qual, newids, "-"));

    gapsToAdd--;
  }
}


void SliceZipper::handleRefIntraGaps(SliceData * reference,
                                     const string & newids,
                                     const char * refconsensus,
                                     long gindex,
                                     long coordinate)
{
  VectorSlice * querySlice = NULL;
  int gapsToAdd = getGapCount(true, gindex, coordinate);

  // If query empties out, don't add erroneous intragaps
  while (gapsToAdd && !m_sliceResults.empty())
  {
    string nuc, qual;
    createGapValues(reference, nuc, qual);

    querySlice = *m_sliceResults.begin();
    m_sliceResults.pop_front();
    // assert (querySlice);

    if (m_promoteAlignmentGaps && querySlice->getConsensus() == '-')
    {
      if (!querySlice->isIntraGap())
      {
        // If there are alignment gaps in the query, 
        // autopromote them to the reference
        gapsToAdd++;
      }
      else
      {
        // intra gap in both???
      }
    }

    if (!querySlice->isIntraGap()) 
    { 
      m_currentQueryPosition++; 
      if (querySlice->getConsensus() != '-') { m_currentQueryUngapped++; }
    }

    mergeSlice(querySlice, nuc, qual, newids, '-', true);
    printSlice(querySlice);

    gapsToAdd--;
  }
}

void SliceZipper::handleSliceMerge(SliceData * reference,
                                   const string & newids,
                                   const char * refconsensus)
{
  if (m_currentReferencePosition == 0)
  {
    handleRefIntraGaps(reference, newids, refconsensus, -1, -1);
  }

  // In merge region, !m_sliceResults.empty()
  VectorSlice * querySlice = *m_sliceResults.begin();
  // assert(querySlice != NULL); 
  
  
  if (m_promoteAlignmentGaps &&
      querySlice->getConsensus() == '-' &&
      refconsensus[0] != '-')
  {
    m_sliceResults.pop_front();

    // If this is an intra gap, don't loop
    // Otherwise autopromote query alignment gaps into the reference
    while (querySlice &&
           !querySlice->isIntraGap() &&
           querySlice->getConsensus() == '-')
    {
      string nuc, qual;
      createGapValues(reference, nuc, qual);

      m_currentQueryPosition++;
      // Don't advance ungapped

      // Print merged reference gaps slices until the query 
      // consensus is non gap
      mergeSlice(querySlice, nuc, qual, newids, '-', true); 
      printSlice(querySlice);

      if (m_sliceResults.empty())
      {
        querySlice = NULL;
      }
      else
      {
        querySlice = *m_sliceResults.begin();
        m_sliceResults.pop_front();
      }
    }
    
    // Merge reference with the query 
    if (querySlice)
    {
      if (!querySlice->isIntraGap()) 
      {
        m_currentQueryPosition++; 
        m_currentQueryUngapped++; // must be nongap consensus
      }

      // Still in merge region consensus match in 'gappedness'
      // or with an intra-gap in the query
      mergeSlice(querySlice,
                 reference->getNucData(),
                 reference->getQualValData(),
                 newids,
                 refconsensus[0],
                 false);

      printSlice(querySlice);
    }
    else
    {
      // Reference right flanks query
      VectorSlice * newSlice;
      newSlice = createVectorSlice(false,
                                   false,
                                   reference->getNucData(),
                                   reference->getQualValData(),
                                   newids,
                                   refconsensus);
      printSlice(newSlice);
    }
  }
  else if (m_promoteAlignmentGaps &&
           refconsensus[0] == '-' &&
           (querySlice->getConsensus() != '-' ||
            querySlice->isIntraGap()))
          
  {
    // Promote a single gap slice into the query, then merge the slice
    // Use the right flanking as the query example
    // Don't pop the head since we will need that element for the next pass
    // Check for intragapness so query alignment gaps are autopromoted
    m_log << "Creating dummy query near "
          << m_currentQueryPosition << endl;

    string nuc, qual;
    createGapValues(querySlice, nuc, qual);

    // Create a gap query slice to merge with
    VectorSlice * newQuerySlice; 
    newQuerySlice = createVectorSlice(false, false, nuc, qual,
                                      querySlice->getReadIdData(),
                                      refconsensus); // refconsensus == '-'

    mergeSlice(newQuerySlice,
               reference->getNucData(),
               reference->getQualValData(),
               newids,
               refconsensus[0],
               true); // '-'

    printSlice(newQuerySlice);
  }
  else // Consensus matches in 'gappedness', or not promoting alignment gaps
  {
    m_sliceResults.pop_front();

    if (!querySlice->isIntraGap()) 
    { 
      m_currentQueryPosition++;
      if (querySlice->getConsensus() != '-') { m_currentQueryUngapped++; }
    }
    
    mergeSlice(querySlice,
               reference->getNucData(),
               reference->getQualValData(),
               newids,
               refconsensus[0],
               false);

    printSlice(querySlice);
  }

  // Check to see if there are any intra gaps to add in the reference
  // after this reference position
  int coordinate = m_currentReferenceUngapped;
  if (coordinate && refconsensus[0] == '-') { coordinate--; }

  handleRefIntraGaps(reference, newids, refconsensus, 
                     m_currentReferencePosition, coordinate);

  m_currentReferencePosition++;
  if (refconsensus[0] != '-') { m_currentReferenceUngapped++; }
}

void SliceZipper::addSlice(SliceData * slice)
{
  char consensus[2] = "\0";
  consensus[0] = slice->getConsensus();

  string newids = m_readMerger.renumberIds(slice);

  if (m_testReferenceGaps && consensus[0] != '-')
  {
    // It's not a gap, we should start merging
    // Otherwise, leave test reference gaps turned on
    m_testReferenceGaps = false;
  }

  if (m_rangeCount == 1)
  {
    // Processing the query
    handleQuerySlice(slice, newids, consensus);
  }
  else if ((m_alignStart < 0) || m_sliceResults.empty() || m_testReferenceGaps) 
  {
    // The reference slices out-flanks query (left or right)
    // Intragaps are meaningless
    
    VectorSlice * newSlice = createVectorSlice(false, 
                                               false,
                                               slice->getNucData(),
                                               slice->getQualValData(),
                                               newids,
                                               consensus);
    printSlice(newSlice);
    m_currentReferencePosition++;
    if (consensus[0] != '-') { m_currentReferenceUngapped++; }

    // Check for left flank
    if (m_alignStart < 0) 
    {
      if (!m_ungappedCoordinates)
      {
        m_alignStart++; 
      }
      else if (consensus[0] != '-') 
      { 
        // Only adjust the offset if the consensus is not a gap
        m_alignStart++; 

        if (m_alignStart == 0)
        {
          // Even though we have printed all of the flanking *bases*
          // it's possible the next few slices have gap as the consensus
          m_testReferenceGaps = true;
        }
      }
    }
    
    // m_testReferenceGaps iff m_alignStart == 0, leave it as is
  }
  else
  {
    // Merge the query and the reference
    handleSliceMerge(slice, newids, consensus);
    m_mergedSlices = true;
  }

  // In all cases it is safe to delete since all saved/printed slices are
  // renumbered for readid
  delete slice;
}

int SliceZipper::endSlices()
{
  if (!m_mergedSlices)
  {
    m_log.fatal("Invalid alignment offset! Type 4");
    throw Exception("Invalid alignment offset!", "SliceZipper");
  }

  while (!m_sliceResults.empty())
  {
    // Query slices right flank merge region
    VectorSlice * querySlice = *m_sliceResults.begin();
    m_sliceResults.pop_front();

    // Intragaps are meaningless, but they may have erroneously been added
    if (!querySlice->isIntraGap()) 
    { 
      m_currentQueryPosition++; 
      if (querySlice->getConsensus() != '-') { m_currentQueryUngapped++; }
    }

    printSlice(querySlice);
  }

  m_out << "  </SliceRange>" << endl;
  m_out << "</Request>"      << endl;

  return m_errorCount;
}

void SliceZipper::handleOptions(Foundation * tf)
{
  tf->getOptions()->addOptionResult("O|alignmentoffset=i", &m_alignStart,
                                    "Specify alignment offset from query to reference\n           negative means reference left flanks query (always 0-based)");

  tf->getOptions()->addOptionResult("q|querygaps=s", &m_queryGaps,
                                    "Specify gaps in query slices");

  tf->getOptions()->addOptionResult("r|referencegaps=s", &m_referenceGaps,
                                    "Specify gaps in reference slices");

  tf->getOptions()->addOptionResult("p|promoteAlignmentGaps!", &m_promoteAlignmentGaps,
                                    "Toggle if alignment gaps should automatically be promoted");

  tf->getOptions()->addOptionResult("R|reversequery!", &m_reverseQuery,
                                    "Query should be reversed before merging");

  tf->getOptions()->addOptionResult("U|ungapped!", &m_ungappedCoordinates,
                                    "Coordinates are all ungapped");

  tf->getOptions()->addOptionResult("1", &m_gapCoordinateOffset,
                                    "Coordinates of gaps are 1-based");

  tf->getOptions()->addOptionResult("refcons", &m_referenceConsensus,
                                    "Use the consensus of the reference for the merged slices");

  RequestManager::handleOptions(tf);

}

void SliceZipper::addGap(int gappos, GapMap & gapmap)
{
  gappos -= m_gapCoordinateOffset;

  m_log << "Adding " << gappos << " (0-based) to gapmap" << endl;

  if (gapmap.find(gappos) != gapmap.end())
  {
    gapmap[gappos]++;
  }
  else
  {
    gapmap[gappos] = 1;
  }
}

void SliceZipper::loadGaps(const string & gaplist, GapMap & gapmap)
{
  // Split the gaplist on ',', load each into the gapmap

  std::string::size_type index = 0;
  std::string::size_type splitIndex;
  int gappos;

  while ((splitIndex = gaplist.find(',', index)) != gaplist.npos)
  {
    gappos = atoi(gaplist.substr(index, splitIndex - index).c_str());
    addGap(gappos, gapmap);
    index = splitIndex + 1;
  }

  gappos = atoi(gaplist.substr(index, 
                               gaplist.length() - index + 1).c_str());
  addGap(gappos, gapmap);
}

VectorSlice::VectorSlice(bool intraGap,
                         bool isGap,
                         const ReadManager & readManager,
                         const char * index, 
                         const char * gindex,
                         const char * existing,
                         long sliceRangeIndex,
                         long sliceRangeGindex)

  : SliceData(readManager,
              index,
              gindex,
              existing,
              sliceRangeIndex,
              sliceRangeGindex)
{
  m_intraGap = intraGap;
  m_isGap = isGap;
}

VectorSlice::~VectorSlice()
{

}

bool VectorSlice::isIntraGap() const
{
  return m_intraGap;
}

bool VectorSlice::isGap() const
{
  return m_isGap;
}

void VectorSlice::setGap(bool)
{
  m_isGap = true;
}
