// $Id: qualityclasses.h,v 1.1.1.1 2005/08/04 14:51:01 mcschatz Exp $

/*! 
 * @page qualityclasses Understanding Quality Classes 
 * @htmlonly <table width="100%"><tr><td align="left" valign="top">@endhtmlonly
 * <b>Draft 'B'</b>
 * @htmlonly</td><td align="right" valign="top">@endhtmlonly
 * $Revision: 1.1.1.1 $
 * @htmlonly</td></tr></table>@endhtmlonly
 *
 * @section qcintro What are Quality Classes?
 * A Quality Class is a single number representation of a set of quality 
 * attributes. They range in value from 0 to 23, and take into account 
 * attributes regarding the quality value, direction, and agreement of the bases
 * in a slice.
 *
 * Typically, slices with quality classes from 1 to 8 are considered high 
 * quality, and 9 through 23 are considered low quality. However, the classes 
 * do not follow a strict ranking overall, so a class may or may not be of 
 * lower quality (by some other measure) than the preceding class. 
 * Low quality slices are typically candidates for resequencing or editing. 
 * Quality class 0 is used to indicate that the slice is in error, or
 * that there was an internal error in the calculation.
 *
 * @htmlonly Quality Class <a href="QualityClasses.ps">Quick Reference Guide</a>.@endhtmlonly
 * <br>
 *
 * @section Definitions
 *
 * @subsection slicedef Slice
 * A slice is a one base wide cut of an assembly. It is composed of 0 or more
 * reads stacked vertically at an offset in the assembly. Each read 
 * contributes a base with a quality value and a direction (either forward or 
 * reverse). A slice also has a consensus, which represents the most likely 
 * base represented by the slice. 
 *
 * @subsection hqdef High/Low Quality Base
 * A base is considered to be of high quality if its quality value is greater
 * than or equal to a high quality threshold. Bases that do not reach the
 * threshold are considered low quality. The threshold can be set as a 
 * parameter of the calculation, but 30 is used by default. The terms high and 
 * low quality reads refer to the reads which contribute the high or low 
 * quality bases respectively.
 *
 * @subsection hqconflictdef High/Low Quality Conflict
 * A slice is said to have a high quality conflict if there is at least one
 * base in the slice that has high quality that does not match the consensus.
 * A low quality conflict means that there is at least one low quality base
 * that does not match the consensus.
 * 
 * @subsection qcattrs Quality Class Attributes
 * Note: For the purposes of agreement, ambiguity codes are not resolved into
 * the component reads, i.e. M and A are in conflict.
 *
 * <b>Conflict Attributes</b>
 *
 * <ul>
 * <li>
 * @b ATTR_No_Conflict <br>
 * Indicates all of the bases in the slice are in agreement with the 
 * consensus.
 * </li>
 *
 * <li>
 * @b ATTR_Non_Quality_Conflict <br>
 * Indicates there is at least one low quality base that conflicts with
 * the consensus.
 * </li>
 *
 * <li>
 * @b ATTR_Quality_Conflict <br>
 * Indicates there is at least one high quality base that conflicts with
 * the consensus. <br><br>
 * @note ATTR_No_Conflict, ATTR_Quality_Conflict, and ATTR_Non_Quality_Conflict
 *       are mutually exclusive and exhaustive for represening if the bases in
 *       a slice are in agreement with the consensus.
 * </li>
 * </ul>
 *
 * <b>Quality Attributes</b>
 *
 * <ul>
 * <li>
 * @b ATTR_Quality_Strand <br>
 * Indicates there is at least one high quality base in at least one
 * direction which agrees with the consensus.
 * </li>
 *
 * <li>
 * @b ATTR_Non_Quality_Strand <br>
 * Indicates there are no high quality bases in either direction that agree
 * with the consensus, but there is at least one low quality base that does
 * agree in at least one direction.
 * </li>
 *
 * <li>
 * @b ATTR_Quality_Same_Strand <br>
 * Indicates there are at least two high quality bases in the same direction
 * agreeing with the consensus in at least one direction. <br><br>
 * - Implies ATTR_Quality_Strand
 * </li>
 *
 * <li>
 * @b ATTR_Non_Quality_Same_Strand <br>
 * Indicates there is at least two bases in the same direction agreeing with 
 * the consensus, but at most one high quality base in that direction. <br><br>
 * - Implies ATTR_Quality_Strand or ATTR_Non_Quality_Strand
 * </li>
 *
 * <li>
 * @b ATTR_Any_Same_Strand <br>
 * Indicates either ATTR_Non_Quality_Same_Strand or ATTR_Quality_Same_Strand.
 * </li>
 *
 * <li>
 * @b ATTR_Quality_Other_Strand <br>
 * Indicates there is at least one high quality base in both directions which
 * agree with the consensus. <br><br>
 * - Implies ATTR_Quality_Strand
 * </li>
 *
 * <li>
 * @b ATTR_Non_Quality_Other_Strand <br>
 * Indicates there is at least one low quality base in both directions which
 * agree with the consensus. <br><br>
 * - Implies either ATTR_Quality_Strand or ATTR_Non_Quality_Strand
 * </li>
 * 
 *
 * <li>
 * @b ATTR_Single_Coverage <br>
 * Indicates that there is only one base that agrees with the consensus.
 * </li>
 * </ul>
 *
 * <b>%Consensus Attributes</b>
 *
 * <ul>
 *
 * <li>
 * @b ATTR_Ambiguity <br>
 * Indicates that the consensus is ambiguous, i.e. not one of {A,C,G,T,-}.
 * </li>
 *
 * <li>
 * @b ATTR_GAP_Consensus <br>
 * Indicates that the consensus is a gap.
 * </li>
 * </ul>
 *
 * <b>%Other Attributes</b>
 *
 * <ul>
 *
 * <li>
 * @b ATTR_Zero_Coverage <br>
 * Indicates that the depth of coverage is zero, there are no underlying
 * sequences in the slice at the current contig position.
 * </li>
 *
 * </ul>
 *
 * <br>
 *
 * @section qcdefs Quality Classes Definitions
 * Once all of the attributes have been calculated, a quality class is
 * choosen to represent the set of attributes present in the slice. Here each
 * class is described and the attributes that it represents are listed and an
 * example slice of that quality class. A slice takes the lowest numbered
 * (highest quality) class that matches the slice in the case where a slice
 * matches more than one class.
 *
 * @note "In agreement" means the base is in agreement with the consensus.
 *
 * <br>
 *
 * @subsection qc0 Quality Class 0
 * A quality class of 0 is used to indicate that there are no reads in the
 * slice, and there is a zero depth of coverage.
 *
 * @note Quality Class 0 is also used to indicate that there was an internal
 * error while calculating quality classes. Please report all slices that
 * do not have gap as consensus and have quality class 0 to bits.closure.
 *
 * - ATTR_Zero_Coverage
 *
 * @verbatim

  Consensus: -   
  ()
 
 

  @endverbatim
 *
 * <br>
 *
 * @subsection qc1 Quality Class 1
 * No conflict, and at least one high quality base in both directions
 * in agreement.
 *
 * - ATTR_No_Conflict
 * - ATTR_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc2 Quality Class 2
 * No conflict, and at least one high quality base in one direction in
 * agreement, and at least one low quality base in the other direction in
 * agreement.
 *
 * - ATTR_No_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc3 Quality Class 3
 * No conflict, and at least one high quality base in agreement, and at least
 * one other base in agreement in that same direction.
 *
 * - ATTR_No_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Any_Same_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 20, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc4 Quality Class 4
 * No conflict, and at least one low quality base in both directions in 
 * agreement.
 *
 * - ATTR_No_Conflict
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 20, Forward)
  (A, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc5 Quality Class 5
 * No conflict, and at least two low quality bases in a single direction in 
 * agreement.
 *
 * - ATTR_No_Conflict
 * - ATTR_Non_Quality_Same_Strand
 *
 * @verbatim

  Consensus: A
  (A, 20, Forward)
  (A, 20, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc6 Quality Class 6
 * Low quality conflict, and at least one high quality base in both directions
 * in agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 30, Reverse)
  (T, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc7 Quality Class 7
 * Low quality conflict, and at least one high quality base in one direction
 * in agreement, and at least one low quality base in the other direction in
 * agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 20, Reverse)
  (T, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc8 Quality Class 8
 * Low quality conflict, and at least one high quality base in one direction
 * in agreement, and at least one other base also in that direction in
 * agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Any_Same_Strand
 *
 * @verbatim

  Consensus: A    
  (A, 30, Forward)
  (A, 20, Forward)
  (T, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc9 Quality Class 9
 * No conflict, but with only a single base in agreement with the consensus.
 *
 * @note Implies the depth of coverage = 1
 *
 * - ATTR_No_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc10 Quality Class 10
 * Low quality conflict, and at least one low quality base in both directions
 * in agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 20, Forward)
  (A, 20, Reverse)
  (T, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc11 Quality Class 11
 * No Conflict, but with only a single base in agreement.
 *
 * @note Implies the depth of coverage = 1
 *
 * - ATTR_No_Conflict
 * - ATTR_Non_Quality_Strand
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: A
  (A, 20, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc12 Quality Class 12
 * Low quality conflict, and at least two low quality bases in the same 
 * direction in agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Non_Quality_Same_Strand
 *
 * @verbatim

  Consensus: A
  (A, 20, Forward)
  (A, 20, Forward)
  (T, 20, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc13 Quality Class 13
 * Low quailty conflict, with only a single high quality base in agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: A
  (A, 35, Forward)
  (T, 10, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc14 Quality Class 14
 * Low quality conflict, with only a single low quality base in agreement.
 *
 * - ATTR_Non_Quality_Conflict
 * - ATTR_Non_Quality_Strand
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: A
  (A, 25, Forward)
  (T, 10, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc15 Quality Class 15
 * High quality conflict, and at least one high quality base in both directions
 * in agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 30, Reverse)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc16 Quality Class 16
 * High quality conflict, and at least one high quality base in one direction
 * in agreement, and at least one low quality base in the other direction in
 * agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 30, Forward)
  (A, 20, Reverse)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc17 Quality Class 17
 * High quality conflict, and at least one high quality base in one direction
 * in agreement, and some other base in the same direction in agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Any_Same_Strand
 *
 * @verbatim

  Consensus: A    
  (A, 30, Forward)
  (A, 20, Forward)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc18 Quality Class 18
 * High quality conflict, and at least one low quality base in both directions
 * in agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Non_Quality_Other_Strand
 *
 * @verbatim

  Consensus: A
  (A, 25, Forward)
  (A, 25, Reverse)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc19 Quality Class 19
 * High quality conflict, and at least two low quality bases in the same
 * direction in agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Non_Quality_Same_Strand
 *
 * @verbatim

  Consensus: A
  (A, 25, Forward)
  (A, 25, Forward)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc20 Quality Class 20
 * High quality conflict, but with only a single high quality base in
 * agreement.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Quality_Strand
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: A
  (A, 40, Forward)
  (T, 30, Reverse)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc21 Quality Class 21
 * %Consensus is ambiguous, and there is only a single base in agreement.
 *
 * - ATTR_Ambiguity
 * - ATTR_Single_Coverage
 *
 * @verbatim

  Consensus: M
  (M, 30, Forward)
  (A, 30, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc22 Quality Class 22
 * %Consensus is ambiguous.
 * 
 * - ATTR_Ambiguity
 *
 * @verbatim

  Consensus: M
  (A, 30, Forward)
  (C, 30, Forward)
  @endverbatim
 *
 * <br>
 *
 * @subsection qc23 Quality Class 23
 * High quality conflict, and only one low quality read in agreement.
 *
 * @note This is also the quality class used to indicate there are no reads 
 * that support that consensus. This can be an indication of a miscalled 
 * consensus.
 *
 * - ATTR_Quality_Conflict
 * - ATTR_Non_Quality_Same_Strand
 * - ATTR_Single_Coverage
 *
 * OR:
 *
 * - Not ATTR_Quality_Strand
 * - Not ATTR_Non_Quality_Strand
 *
 * @verbatim

  Consensus: A          Consensus: A
  (A, 20, Forward)      (C, 30, Forward)
  (C, 30, Forward)      (C, 30, Reverse)
  @endverbatim
 *
 * @htmlonly<hr><p align=right>$Date: 2005/08/04 14:51:01 $</p>@endhtmlonly
 */

